/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file audiod.c the paraslash's audio daemon */
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>

#include "para.h"
#include "error.h"
#include "audiod.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "recv.h"
#include "filter.h"
#include "grab_client.cmdline.h"
#include "grab_client.h"
#include "client.cmdline.h"
#include "client.h"
#include "audiod.h"
#include "net.h"
#include "daemon.h"
#include "string.h"
#include "fd.h"
#include "write.h"
#include "write_common.h"
#include "signal.h"

/** define the array of error lists needed by para_audiod */
INIT_AUDIOD_ERRLISTS;
/** define the array containing all supported audio formats */
const char *audio_formats[] = {AUDIOD_AUDIO_FORMAT_ARRAY NULL};

/** Defines how audiod handles one supported audio format. */
struct audio_format_info {
	/** pointer to the receiver for this audio format */
	struct receiver *receiver;
	/** the receiver configuration */
	void *receiver_conf;
	/** the number of filters that should be activated for this audio format */
	unsigned int num_filters;
	/** Array of filter numbers to be activated. */
	unsigned *filter_nums;
	/** Pointer to the array of filter configurations. */
	void **filter_conf;
	/** the number of filters that should be activated for this audio format */
	unsigned int num_writers;
	/** Array of writer numbers to be activated. */
	int *writer_nums;
	/** pointer to the array of writer configurations */
	void **writer_conf;
	/** do not start receiver/filters/writer before this time */
	struct timeval restart_barrier;
};

/**
 * para_audiod uses \p MAX_STREAM_SLOTS different slots, each of which may
 * be associated with a receiver/filter/writer triple. This array holds all
 * information on the status of these slots.
 *
 * \sa struct slot_info
 * */
struct slot_info slot[MAX_STREAM_SLOTS];

/** The vss status flags audiod is interested in. */
enum vss_status_flags {
	/** Whether the 'N' flag is set. */
	VSS_STATUS_FLAG_NEXT = 1,
	/** The 'P' flag is set. */
	VSS_STATUS_FLAG_PLAYING = 2,
};

/**
 * The task for obtaining para_server's status (para_client stat).
 *
 * \sa struct task, struct sched.
 */
struct status_task {
	/** The associated task structure of audiod. */
	struct task task;
	/** Client data associated with the stat task. */
	struct client_task *ct;
	/** Do not restart client command until this time. */
	struct timeval restart_barrier;
	/** Last time we received status data from para_server. */
	struct timeval last_status_read;
	/** The offset value announced by para_server. */
	int offset_seconds;
	/** The length of the current audio file as announced by para_server. */
	int length_seconds;
	/** The start of the current stream from the view of para_server. */
	struct timeval server_stream_start;
	/** The average time deviation between para_server and para_audiod. */
	struct timeval sa_time_diff;
	/** Whether client time is ahead of server time. */
	int sa_time_diff_sign;
	/** The 'P' and the 'N' flags as announced by para_server. */
	enum vss_status_flags vss_status;
	/** Number of times the clock difference is to be checked. */
	unsigned clock_diff_count;
	/** When to start the next check for clock difference. */
	struct timeval clock_diff_barrier;
	/** Number of the audio format as announced by para_server. */
	int current_audio_format_num;
};

/** The array of status items sent by para_server. */
char *stat_item_values[NUM_STAT_ITEMS] = {NULL};

/**
 * the current mode of operation of which can be changed by the on/off/cycle
 * commands. It is either, AUDIOD_OFF, AUDIOD_ON or AUDIOD_STANDBY.
 */
int audiod_status = AUDIOD_ON;

/**
 * the gengetopt args_info struct that holds information on all command line
 * arguments
 */
struct audiod_args_info conf;

static char *socket_name;
static struct audio_format_info afi[NUM_AUDIO_FORMATS];

static struct signal_task signal_task_struct, *sig_task = &signal_task_struct;

static struct status_task status_task_struct;

/**
 * the task that calls the status command of para_server
 *
 * \sa struct status_task
 */
static struct status_task *stat_task = &status_task_struct;
static struct timeval initial_delay_barrier;

/**
 * the task for handling audiod commands
 *
 * \sa struct task, struct sched
 */
struct command_task {
	/** the local listening socket */
	int fd;
	/** the associated task structure */
	struct task task;
};

/** iterate over all supported audio formats */
#define FOR_EACH_AUDIO_FORMAT(af) for (af = 0; af < NUM_AUDIO_FORMATS; af++)

/**
 * get the audio format number
 * \param name the name of the audio format
 *
 * \return The audio format number on success, -E_UNSUPPORTED_AUDIO_FORMAT if
 * \a name is not a supported audio format.
 */
int get_audio_format_num(char *name)
{
	int i;

	while (para_isspace(*name))
		name++;
	FOR_EACH_AUDIO_FORMAT(i)
		if (!strcmp(name, audio_formats[i]))
			return i;
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

char *get_time_string(int slot_num)
{
	int ret, seconds = 0, length;
	struct timeval *tmp, sum, sss, /* server stream start */
		wtime, /* now - writer start */
		rskip; /* receiver start - sss */
	struct slot_info *s = slot_num < 0? NULL : &slot[slot_num];

	if (audiod_status == AUDIOD_OFF)
		goto empty;
	if (!(stat_task->vss_status & VSS_STATUS_FLAG_PLAYING)) {
		if (stat_task->length_seconds) /* paused */
			return NULL;
		goto empty; /* stopped */
	}
	if (audiod_status == AUDIOD_ON && !s)
		goto empty;
	/* valid status items and playing */
	if (s) { /* writer active in this slot */
		length = s->seconds_total;
		tmp = &s->server_stream_start;
	} else { /* standby mode, rely on status items */
		length = stat_task->length_seconds;
		tmp = &stat_task->server_stream_start;
	}
	if (stat_task->sa_time_diff_sign > 0)
		tv_diff(tmp, &stat_task->sa_time_diff, &sss);
	else
		tv_add(tmp, &stat_task->sa_time_diff, &sss);
	if (!s) {
		struct timeval diff;
		tv_diff(now, &sss, &diff);
		seconds = diff.tv_sec + stat_task->offset_seconds;
		goto out;
	}
	tv_diff(now, &s->wstime, &wtime);
	seconds = s->offset_seconds;
	ret = tv_diff(&s->rstime, &sss, &rskip);
	if (ret > 0) { /* audiod was started in the middle of the stream */
		tv_add(&wtime, &rskip, &sum);
		seconds += sum.tv_sec;
	} else
		seconds += wtime.tv_sec;
out:
	seconds = PARA_MIN(seconds, length);
	seconds = PARA_MAX(seconds, 0);
	return make_message(
		"%s: %s%d:%02d [%d:%02d] (%d%%/%d:%02d)\n",
		status_item_list[SI_PLAY_TIME],
		s? "" : "~",
		seconds / 60,
		seconds % 60,
		(length - seconds) / 60,
		(length - seconds) % 60,
		length? (seconds * 100 + length / 2) / length : 0,
		length / 60,
		length % 60
	);
empty:
	return make_message("%s:\n", status_item_list[SI_PLAY_TIME]);
}

static int want_colors(void)
{
	if (conf.color_arg == color_arg_no)
		return 0;
	if (conf.color_arg == color_arg_yes)
		return 1;
	if (conf.logfile_given)
		return 0;
	return isatty(STDERR_FILENO);
}

static void parse_config_or_die(void)
{
	int ret;
	char *config_file;
	struct audiod_cmdline_parser_params params = {
		.override = 0,
		.initialize = 0,
		.check_required = 1,
		.check_ambiguity = 0,
		.print_errors = 1
	};

	if (conf.config_file_given)
		config_file = para_strdup(conf.config_file_arg);
	else {
		char *home = para_homedir();
		config_file = make_message("%s/.paraslash/audiod.conf", home);
		free(home);
	}
	ret = file_exists(config_file);
	if (conf.config_file_given && !ret) {
		PARA_EMERG_LOG("can not read config file %s\n", config_file);
		goto err;
	}
	if (ret)
		audiod_cmdline_parser_config_file(config_file, &conf, &params);
	free(config_file);
	daemon_set_loglevel(conf.loglevel_arg);
	return;
err:
	free(config_file);
	exit(EXIT_FAILURE);
}

static void setup_signal_handling(void)
{
	sig_task->fd = para_signal_init();
	PARA_INFO_LOG("signal pipe: fd %d\n", sig_task->fd);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGHUP);
	para_sigaction(SIGPIPE, SIG_IGN);
}

static void clear_slot(int slot_num)
{
	struct slot_info *s = &slot[slot_num];

	PARA_INFO_LOG("clearing slot %d\n", slot_num);
	memset(s, 0, sizeof(struct slot_info));
	s->format = -1;
}

static void close_receiver(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a;

	if (s->format < 0 || !s->receiver_node)
		return;
	a = &afi[s->format];
	PARA_NOTICE_LOG("closing %s receiver in slot %d\n",
		audio_formats[s->format], slot_num);
	a->receiver->close(s->receiver_node);
	free(s->receiver_node);
	s->receiver_node = NULL;
}

static void kill_all_decoders(int error)
{
	int i;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		if (s->wng && s->wng->task.error >= 0) {
			PARA_INFO_LOG("deactivating wng in slot %d\n",
				i);
			s->wng->task.error = error;
		}
		if (s->fc && s->fc->task.error >= 0) {
			PARA_INFO_LOG("deactivatimg filter chain in slot %d\n", i);
			s->fc->task.error = error;
		}
		if (s->receiver_node && s->receiver_node->task.error >= 0) {
			PARA_INFO_LOG("deactivating receiver_node in slot %d\n", i);
			s->receiver_node->task.error = error;
		}
	}
}

static int get_empty_slot(void)
{
	int i;
	struct slot_info *s;

	FOR_EACH_SLOT(i) {
		s = &slot[i];
		if (s->format < 0) {
			clear_slot(i);
			return i;
		}
		if (s->wng || s->receiver_node || s->fc)
			continue;
		clear_slot(i);
		return i;
	}
	return -E_NO_MORE_SLOTS;
}

/**
 * get the number of filters
 *
 * \param audio_format_num the number identifying the audio format
 *
 * \return the number of filters for the given audio format
 *
 * \sa struct filter;
 */
int num_filters(int audio_format_num)
{
	return afi[audio_format_num].num_filters;
}

static void open_filters(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = &afi[s->format];
	struct filter_node *fn;
	int nf = a->num_filters;
	int i;

	s->fc = NULL;
	if (!nf)
		return;
	PARA_INFO_LOG("opening %s filters\n", audio_formats[s->format]);
	s->fc = para_calloc(sizeof(struct filter_chain));
	s->fc->filter_nodes = para_malloc(nf * sizeof(struct filter_node));
	s->fc->inbufp = &s->receiver_node->buf;
	s->fc->in_loaded = &s->receiver_node->loaded;
	s->fc->input_error = &s->receiver_node->task.error;
	s->fc->task.pre_select = NULL;
	s->fc->task.post_select = filter_post_select;
	s->fc->task.error = 0;
	s->fc->num_filters = nf;

	s->receiver_node->output_error = &s->fc->task.error;
	sprintf(s->fc->task.status, "filter chain");
	FOR_EACH_FILTER_NODE(fn, s->fc, i) {
		struct filter *f = filters + a->filter_nums[i];
		fn->filter_num = a->filter_nums[i];
		fn->conf = a->filter_conf[i];
		fn->fc = s->fc;
		fn->loaded = 0;
		INIT_LIST_HEAD(&fn->callbacks);
		f->open(fn);
		PARA_NOTICE_LOG("%s filter %d/%d (%s) started in slot %d\n",
			audio_formats[s->format], i,  nf, f->name, slot_num);
		s->fc->outbufp = &fn->buf;
		s->fc->out_loaded = &fn->loaded;
	}
	register_task(&s->fc->task);
}

static void open_writers(int slot_num)
{
	int ret, i;
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = &afi[s->format];

	PARA_INFO_LOG("opening %s writers\n", audio_formats[s->format]);
	if (!a->num_writers)
		s->wng = setup_default_wng();
	else
		s->wng = wng_new(a->num_writers);
	if (s->fc) {
		s->wng->bufp = s->fc->outbufp;
		s->wng->loaded = s->fc->out_loaded;
		s->wng->input_error = &s->fc->task.error;
		s->wng->channels = &s->fc->channels;
		s->wng->samplerate = &s->fc->samplerate;
		s->fc->output_error = &s->wng->task.error;
		PARA_INFO_LOG("samplerate: %d\n", *s->wng->samplerate);
	} else {
		s->wng->bufp = &s->receiver_node->buf;
		s->wng->loaded = &s->receiver_node->loaded;
		s->wng->input_error = &s->receiver_node->task.error;
	}
	for (i = 0; i < a->num_writers; i++) {
		s->wng->writer_nodes[i].conf = a->writer_conf[i];
		s->wng->writer_nodes[i].writer_num = a->writer_nums[i];
	}
	ret = wng_open(s->wng);
	if (ret < 0)
		return;
	s->wstime = *now;
	s->server_stream_start = stat_task->server_stream_start.tv_sec?
		stat_task->server_stream_start : *now;
	s->offset_seconds = stat_task->offset_seconds;
	s->seconds_total = stat_task->length_seconds;
	activate_inactive_grab_clients(slot_num, s->format, s->fc);
}

static int open_receiver(int format)
{
	struct audio_format_info *a = &afi[format];
	struct slot_info *s;
	int ret, slot_num;
	struct receiver_node *rn;
	const struct timeval restart_delay = {2, 0};

	ret = get_empty_slot();
	if (ret < 0)
		goto err;
	slot_num = ret;
	s = &slot[slot_num];
	s->format = format;
	s->receiver_node = para_calloc(sizeof(struct receiver_node));
	rn = s->receiver_node;
	rn->receiver = a->receiver;
	rn->conf = a->receiver_conf;
	ret = a->receiver->open(s->receiver_node);
	if (ret < 0) {
		free(s->receiver_node);
		s->receiver_node = NULL;
		goto err;
	}
	PARA_NOTICE_LOG("started %s: %s receiver in slot %d\n",
		audio_formats[s->format], a->receiver->name, slot_num);
	rn->task.pre_select = a->receiver->pre_select;
	rn->task.post_select = a->receiver->post_select;
	s->rstime = *now;
	sprintf(rn->task.status, "%s receiver node", rn->receiver->name);
	register_task(&rn->task);
	ret = 1;
err:
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	tv_add(now, &restart_delay, &afi[format].restart_barrier);
	return ret;
}

/* return: 0: Not running, 1: Running, -1: Running but eof (or error) */
static int receiver_running(int format)
{
	int i, ret = 0;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		if (s->format != format)
			continue;
		if (!s->receiver_node)
			continue;
		if (s->receiver_node->task.error >= 0)
			return 1;
		ret = -1;
	}
	return ret;
}

static void open_current_receiver(struct sched *s)
{
	struct timeval diff;
	int ret, cafn = stat_task->current_audio_format_num;

	if (cafn < 0 || !stat_task->ct)
		return;
	/* Do nothing if the 'N' flag is set or the 'P' flag is unset */
	if (stat_task->vss_status != VSS_STATUS_FLAG_PLAYING)
		return;
	ret = receiver_running(cafn);
	if (ret > 0) /* already running and not eof */
		return;
	if (ret < 0) { /* eof */
		/*
		 * para_server uses a zero start time during the announcement
		 * period, i.e. before it sends the first chunk. Wait until
		 * this period begins to avoid restarting the receiver that
		 * belongs to the file just completed.
		 */
		if (stat_task->server_stream_start.tv_sec)
			return;
	}
	if (tv_diff(now, &afi[cafn].restart_barrier, &diff) < 0) {
		/* avoid busy loop */
		s->timeout = diff;
		return;
	}
	/* start a new receiver */
	open_receiver(cafn);
}

static unsigned compute_time_diff(const struct timeval *status_time)
{
	struct timeval tmp, diff;
	static unsigned count;
	int sign, sa_time_diff_sign = stat_task->sa_time_diff_sign;
	const struct timeval max_deviation = {0, 500 * 1000};
	const int time_smooth = 5;

	if (!status_time)
		return count;
	sign = tv_diff(status_time, now, &diff);
//		PARA_NOTICE_LOG("%s: sign = %i, sa_time_diff_sign = %i\n", __func__,
//			sign, sa_time_diff_sign);
	if (!count) {
		sa_time_diff_sign = sign;
		stat_task->sa_time_diff = diff;
		count++;
		goto out;
	}
	if (count > 5) {
		int s = tv_diff(&diff, &stat_task->sa_time_diff, &tmp);
		if (tv_diff(&max_deviation, &tmp, NULL) < 0)
			PARA_WARNING_LOG("time diff jump: %lims\n",
				s * tv2ms(&tmp));
	}
	count++;
	sa_time_diff_sign = tv_convex_combination(
		sa_time_diff_sign * time_smooth, &stat_task->sa_time_diff,
		count > 10? sign : sign * time_smooth, &diff,
		&tmp);
	stat_task->sa_time_diff = tmp;
	PARA_INFO_LOG("time diff (cur/avg): %s%lums/%s%lums\n",
		sign > 0? "+" : "-",
		tv2ms(&diff),
		sa_time_diff_sign ? "+" : "-",
		tv2ms(&stat_task->sa_time_diff)
	);
out:
	stat_task->sa_time_diff_sign = sa_time_diff_sign;
	return count;
}

static int check_stat_line(char *line, __a_unused void *data)
{
	int itemnum;
	size_t ilen = 0;
	long unsigned sec, usec;
	char *tmp;

	//PARA_INFO_LOG("line: %s\n", line);
	if (!line)
		return 1;
	itemnum = stat_line_valid(line);
	if (itemnum < 0) {
		PARA_WARNING_LOG("invalid status line: %s\n", line);
		return 1;
	}
	if (stat_task->clock_diff_count && itemnum != SI_CURRENT_TIME)
		return 1;
	tmp = make_message("%s\n", line);
	stat_client_write(tmp, itemnum);
	free(tmp);
	free(stat_item_values[itemnum]);
	stat_item_values[itemnum] = para_strdup(line);
	ilen = strlen(status_item_list[itemnum]);
	switch (itemnum) {
	case SI_STATUS_FLAGS:
		stat_task->vss_status = 0;
		if (strchr(line, 'N'))
			stat_task->vss_status |= VSS_STATUS_FLAG_NEXT;
		if (strchr(line, 'P'))
			stat_task->vss_status |= VSS_STATUS_FLAG_PLAYING;
		break;
	case SI_OFFSET:
		stat_task->offset_seconds = atoi(line + ilen + 1);
		break;
	case SI_SECONDS_TOTAL:
		stat_task->length_seconds = atoi(line + ilen + 1);
		break;
	case SI_STREAM_START:
		if (sscanf(line + ilen + 1, "%lu.%lu", &sec, &usec) == 2) {
			struct timeval a_start, delay;
			delay.tv_sec = conf.stream_delay_arg / 1000;
			delay.tv_usec = (conf.stream_delay_arg % 1000) * 1000;
			stat_task->server_stream_start.tv_sec = sec;
			stat_task->server_stream_start.tv_usec = usec;
			if (compute_time_diff(NULL) > 2) {
				if (stat_task->sa_time_diff_sign < 0)
					tv_add(&stat_task->server_stream_start,
						&stat_task->sa_time_diff, &a_start);
				else
					tv_diff(&stat_task->server_stream_start,
						&stat_task->sa_time_diff, &a_start);
				tv_add(&a_start, &delay, &initial_delay_barrier);
			}
		}
		break;
	case SI_CURRENT_TIME:
		if (sscanf(line + ilen + 1, "%lu.%lu", &sec, &usec) == 2) {
			struct timeval tv = {sec, usec};
			compute_time_diff(&tv);
		}
		break;
	case SI_FORMAT:
		stat_task->current_audio_format_num = get_audio_format_num(
			line + ilen + 1);
	}
	return 1;
}

static int parse_stream_command(const char *txt, char **cmd)
{
	char *p = strchr(txt, ':');
	int i;

	if (!p)
		return -E_MISSING_COLON;
	p++;
	FOR_EACH_AUDIO_FORMAT(i) {
		if (strncmp(txt, audio_formats[i], strlen(audio_formats[i])))
			continue;
		*cmd = p;
		return i;
	}
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

static int add_filter(int format, char *cmdline)
{
	struct audio_format_info *a = &afi[format];
	int filter_num, nf = a->num_filters;

	filter_num = check_filter_arg(cmdline, &a->filter_conf[nf]);
	if (filter_num < 0)
		return filter_num;
	a->filter_nums[nf] = filter_num;
	a->num_filters++;
	PARA_INFO_LOG("%s filter %d: %s\n", audio_formats[format], nf,
		filters[filter_num].name);
	return filter_num;
}

static int parse_writer_args(void)
{
	int i, ret, nw;
	char *cmd;
	struct audio_format_info *a;

	nw = PARA_MAX(1U, conf.writer_given);
	PARA_INFO_LOG("maximal number of writers: %d\n", nw);
	FOR_EACH_AUDIO_FORMAT(i) {
		a = &afi[i];
		a->writer_conf = para_malloc(nw * sizeof(void *));
		a->writer_nums = para_malloc(nw * sizeof(int));
		a->num_writers = 0;
	}
	for (i = 0; i < conf.writer_given; i++) {
		void *wconf;
		int writer_num;
		ret = parse_stream_command(conf.writer_arg[i], &cmd);
		if (ret < 0)
			goto out;
		a = &afi[ret];
		nw = a->num_writers;
		wconf = check_writer_arg(cmd, &writer_num);
		if (!wconf) {
			ret = writer_num;
			goto out;
		}
		a->writer_nums[nw] = writer_num;
		a->writer_conf[nw] = wconf;
		PARA_INFO_LOG("%s writer #%d: %s\n", audio_formats[ret],
			nw, writer_names[writer_num]);
		a->num_writers++;
	}
	ret = 1;
out:
	return ret;
}

static int parse_receiver_args(void)
{
	int i, ret, receiver_num;
	char *cmd = NULL;
	struct audio_format_info *a;

	for (i = conf.receiver_given - 1; i >= 0; i--) {
		char *arg = conf.receiver_arg[i];
		char *recv_arg = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!recv_arg)
			goto out;
		*recv_arg = '\0';
		recv_arg++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		afi[ret].receiver_conf = check_receiver_arg(recv_arg, &receiver_num);
		if (!afi[ret].receiver_conf) {
			ret = -E_RECV_SYNTAX;
			goto out;
		}
		afi[ret].receiver = &receivers[receiver_num];
	}
	/* use the first available receiver with no arguments
	 * for those audio formats for which no receiver
	 * was specified
	 */
	cmd = para_strdup(receivers[0].name);
	FOR_EACH_AUDIO_FORMAT(i) {
		a = &afi[i];
		if (a->receiver_conf)
			continue;
		a->receiver_conf = check_receiver_arg(cmd, &receiver_num);
		if (!a->receiver_conf)
			return -E_RECV_SYNTAX;
		a->receiver = &receivers[receiver_num];
	}
	ret = 1;
out:
	free(cmd);
	return ret;
}

static int init_default_filters(void)
{
	int i, ret = 1;

	FOR_EACH_AUDIO_FORMAT(i) {
		struct audio_format_info *a = &afi[i];
		char *tmp;
		int j;

		if (a->num_filters)
			continue; /* no default -- nothing to to */
		/* add "dec" to audio format name */
		tmp = make_message("%sdec", audio_formats[i]);
		for (j = 0; filters[j].name; j++)
			if (!strcmp(tmp, filters[j].name))
				break;
		free(tmp);
		ret = -E_UNSUPPORTED_FILTER;
		if (!filters[j].name)
			goto out;
		tmp = para_strdup(filters[j].name);
		ret = add_filter(i, tmp);
		free(tmp);
		if (ret < 0)
			goto out;
		PARA_INFO_LOG("%s -> default filter: %s\n", audio_formats[i],
			filters[j].name);
	}
out:
	return ret;
}

static int parse_filter_args(void)
{
	int i, ret, nf;

	nf = PARA_MAX(1U, conf.filter_given);
	PARA_INFO_LOG("maximal number of filters: %d\n", nf);
	FOR_EACH_AUDIO_FORMAT(i) {
		afi[i].filter_conf = para_malloc(nf * sizeof(void *));
		afi[i].filter_nums = para_malloc(nf * sizeof(unsigned));
	}
	if (!conf.no_default_filters_given)
		return init_default_filters();
	for (i = 0; i < conf.filter_given; i++) {
		char *arg = conf.filter_arg[i];
		char *filter_name = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!filter_name)
			goto out;
		*filter_name = '\0';
		filter_name++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		ret = add_filter(ret, filter_name);
		if (ret < 0)
			goto out;
	}
	ret = init_default_filters(); /* use default values for the rest */
out:
	return ret;
}

static int parse_stream_args(void)
{
	int ret;

	ret = parse_receiver_args();
	if (ret < 0)
		return ret;
	ret = parse_filter_args();
	if (ret < 0)
		return ret;
	ret = parse_writer_args();
	if (ret < 0)
		return ret;
	return 1;
}

/* does not unlink socket on errors */
static int audiod_get_socket(void)
{
	struct sockaddr_un unix_addr;
	int ret, fd;

	if (conf.socket_given)
		socket_name = para_strdup(conf.socket_arg);
	else {
		char *hn = para_hostname();
		socket_name = make_message("/var/paraslash/audiod_socket.%s",
			hn);
		free(hn);
	}
	PARA_NOTICE_LOG("local socket: %s\n", socket_name);
	if (conf.force_given)
		unlink(socket_name);
	ret = create_local_socket(socket_name, &unix_addr,
		S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IWOTH);
	if (ret < 0)
		goto err;
	fd = ret;
	if (listen(fd , 5) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		goto err;
	return fd;
err:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

static void signal_pre_select(struct sched *s, struct task *t)
{
	struct signal_task *st = container_of(t, struct signal_task, task);
	para_fd_set(st->fd, &s->rfds, &s->max_fileno);
}

static void signal_post_select(struct sched *s, struct task *t)
{
	struct signal_task *st = container_of(t, struct signal_task, task);

	if (!FD_ISSET(st->fd, &s->rfds))
		return;

	st->signum = para_next_signal();
	switch (st->signum) {
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		PARA_EMERG_LOG("terminating on signal %d\n", st->signum);
		clean_exit(EXIT_FAILURE, "caught deadly signal");
	}
}

static void signal_setup_default(struct signal_task *st)
{
	st->task.pre_select = signal_pre_select;
	st->task.post_select = signal_post_select;
	sprintf(st->task.status, "signal task");
}

static void command_pre_select(struct sched *s, struct task *t)
{
	struct command_task *ct = container_of(t, struct command_task, task);
	para_fd_set(ct->fd, &s->rfds, &s->max_fileno);
}

static void command_post_select(struct sched *s, struct task *t)
{
	int ret;
	struct command_task *ct = container_of(t, struct command_task, task);

	audiod_status_dump();
	if (!FD_ISSET(ct->fd, &s->rfds))
		return;
	ret = handle_connect(ct->fd);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
}

static void init_command_task(struct command_task *ct)
{
	ct->task.pre_select = command_pre_select;
	ct->task.post_select = command_post_select;
	ct->task.error = 0;
	ct->fd = audiod_get_socket(); /* doesn't return on errors */
	sprintf(ct->task.status, "command task");
}

static void close_stat_pipe(void)
{
	int i;

	if (!stat_task->ct)
		return;
	client_close(stat_task->ct);
	stat_task->ct = NULL;
	FOR_EACH_STATUS_ITEM(i) {
		free(stat_item_values[i]);
		stat_item_values[i] = NULL;
	}
	dump_empty_status();
	stat_task->length_seconds = 0;
	stat_task->offset_seconds = 0;
	stat_task->vss_status = 0;
	audiod_status_dump();
}

/**
 * close the connection to para_server and exit
 *
 * \param status the exit status which is passed to exit(3)
 * \param msg the log message
 *
 * Log \a msg with loglevel \p EMERG, close the connection to para_server if
 * open, and call \p exit(status). \a status should be either EXIT_SUCCESS or
 * EXIT_FAILURE.
 *
 * \sa exit(3)
 */
void __noreturn clean_exit(int status, const char *msg)
{
	PARA_EMERG_LOG("%s\n", msg);
	if (socket_name)
		unlink(socket_name);
	close_stat_pipe();
	exit(status);
}

/* avoid busy loop if server is down */
static void set_stat_task_restart_barrier(unsigned seconds)
{
	struct timeval delay = {seconds, 0};
	tv_add(now, &delay, &stat_task->restart_barrier);
}

static void try_to_close_slot(int slot_num)
{
	struct slot_info *s = &slot[slot_num];

	if (s->format < 0)
		return;
	if (s->receiver_node && s->receiver_node->task.error != -E_TASK_UNREGISTERED)
		return;
	if (s->fc && s->fc->task.error != -E_TASK_UNREGISTERED)
		return;
	if (s->wng && s->wng->task.error != -E_TASK_UNREGISTERED)
		return;
	PARA_INFO_LOG("closing slot %d\n", slot_num);
	wng_close(s->wng);
	close_filters(s->fc);
	free(s->fc);
	close_receiver(slot_num);
	clear_slot(slot_num);
}

/*
 * Check if any receivers/filters/writers need to be started and do so if
 * necessary.
 */
static void start_stop_decoders(struct sched *s)
{
	int i;

	FOR_EACH_SLOT(i)
		try_to_close_slot(i);
	if (audiod_status != AUDIOD_ON ||
			!(stat_task->vss_status & VSS_STATUS_FLAG_PLAYING))
		return kill_all_decoders(-E_NOT_PLAYING);
	open_current_receiver(s);
	FOR_EACH_SLOT(i) {
		struct slot_info *sl = &slot[i];
		struct audio_format_info *a;
		struct timeval diff;

		if (sl->format < 0)
			continue;
		a = &afi[sl->format];
		if (!sl->receiver_node)
			continue;
		if ((!a->num_filters || sl->fc) && sl->wng)
			continue; /* everything already started */
		if (!a->num_filters) {
			if (sl->receiver_node->loaded && !sl->wng) {
				open_writers(i);
			}
			continue;
		}
		if (sl->receiver_node->loaded && !sl->fc) {
			open_filters(i);
			continue;
		}
		if (sl->wng || !sl->fc || !*sl->fc->out_loaded)
			continue;
		if (tv_diff(now, &initial_delay_barrier, &diff) > 0) {
			open_writers(i);
			continue;
		}
		PARA_INFO_LOG("initial delay: %lu ms left\n", tv2ms(&diff));
		if (tv_diff(&s->timeout, &diff, NULL) > 0) {
			s->timeout = diff;
		}
	}
}


/* restart the client task if necessary */
static void status_pre_select(struct sched *s, struct task *t)
{
	struct status_task *st = container_of(t, struct status_task, task);

	if (audiod_status == AUDIOD_OFF) {
		if (!st->ct)
			goto out;
		if (st->ct->task.error >= 0) {
			st->ct->task.error = -E_AUDIOD_OFF;
			goto out;
		}
		if (st->ct->task.error != -E_TASK_UNREGISTERED)
			goto out;
		close_stat_pipe();
		st->clock_diff_count = conf.clock_diff_count_arg;
		goto out;
	}
	if (st->ct) {
		unsigned bytes_left;
		if (st->ct->task.error < 0) {
			if (st->ct->task.error != -E_TASK_UNREGISTERED)
				goto out;
			close_stat_pipe();
			goto out;
		}
		if (st->ct->status != CL_RECEIVING)
			goto out;
		bytes_left = for_each_line(st->ct->buf, st->ct->loaded,
			&check_stat_line, NULL);
		if (st->ct->loaded != bytes_left) {
			st->last_status_read = *now;
			st->ct->loaded = bytes_left;
		} else {
			struct timeval diff;
			tv_diff(now, &st->last_status_read, &diff);
			if (diff.tv_sec > 61)
				st->ct->task.error = -E_STATUS_TIMEOUT;
		}
		goto out;
	}
	if (tv_diff(now, &st->restart_barrier, NULL) < 0)
		goto out;
	if (st->clock_diff_count) { /* get status only one time */
		char *argv[] = {"audiod", "stat", "1", NULL};
		int argc = 3;
		PARA_INFO_LOG("clock diff count: %d\n", st->clock_diff_count);
		st->clock_diff_count--;
		client_open(argc, argv, &st->ct, NULL);
		set_stat_task_restart_barrier(2);

	} else {
		char *argv[] = {"audiod", "stat", NULL};
		int argc = 2;
		client_open(argc, argv, &st->ct, NULL);
		set_stat_task_restart_barrier(5);
	}
	free(stat_item_values[SI_BASENAME]);
	stat_item_values[SI_BASENAME] = make_message(
		"%s: no connection to para_server\n",
		status_item_list[SI_BASENAME]);
	stat_client_write(stat_item_values[SI_BASENAME],
		SI_BASENAME);
	st->last_status_read = *now;
out:
	start_stop_decoders(s);
}

static void init_status_task(struct status_task *st)
{
	memset(st, 0, sizeof(struct status_task));
	st->task.pre_select = status_pre_select;
	st->sa_time_diff_sign = 1;
	st->clock_diff_count = conf.clock_diff_count_arg;
	st->current_audio_format_num = -1;
	sprintf(st->task.status, "status task");
}

static void set_initial_status(void)
{
	audiod_status = AUDIOD_ON;
	if (!conf.mode_given)
		return;
	if (!strcmp(conf.mode_arg, "sb")) {
		audiod_status = AUDIOD_STANDBY;
		return;
	}
	if (!strcmp(conf.mode_arg, "off")) {
		audiod_status = AUDIOD_OFF;
		return;
	}
	if (strcmp(conf.mode_arg, "on"))
		PARA_WARNING_LOG("invalid mode\n");
}

__noreturn static void print_help_and_die(void)
{
	int d = conf.detailed_help_given;
	const char **p = d? audiod_args_info_detailed_help
		: audiod_args_info_help;

	printf_or_die("%s\n\n", AUDIOD_CMDLINE_PARSER_PACKAGE "-"
		AUDIOD_CMDLINE_PARSER_VERSION);
	printf_or_die("%s\n\n", audiod_args_info_usage);
	for (; *p; p++)
		printf_or_die("%s\n", *p);
	print_receiver_helps(d);
	print_filter_helps(d);
	print_writer_helps(d);
	exit(0);
}

static void init_colors_or_die(void)
{
	int ret, i;

	if (!want_colors())
		return;
	daemon_set_default_log_colors();
	daemon_set_flag(DF_COLOR_LOG);
	for (i = 0; i < conf.log_color_given; i++) {
		ret = daemon_set_log_color(conf.log_color_arg[i]);
		if (ret < 0)
			exit(EXIT_FAILURE);
	}
}

/**
 * the main function of para_audiod
 *
 * \param argc usual argument count
 * \param argv usual argument vector
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE
 *
 * \sa para_audiod(1)
 * */
int main(int argc, char *argv[])
{
	int ret, i;
	static struct sched s;
	struct command_task command_task_struct, *cmd_task = &command_task_struct;
	struct audiod_cmdline_parser_params params = {
		.override = 0,
		.initialize = 1,
		.check_required = 0,
		.check_ambiguity = 0,
		.print_errors = 1
	};

	valid_fd_012();
	if (audiod_cmdline_parser_ext(argc, argv, &conf, &params))
		exit(EXIT_FAILURE);
	HANDLE_VERSION_FLAG("audiod", conf);
	/* init receivers/filters/writers early to make help work */
	recv_init();
	filter_init();
	writer_init();
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	drop_privileges_or_die(conf.user_arg, conf.group_arg);
	parse_config_or_die();
	init_colors_or_die();
	daemon_set_flag(DF_LOG_TIME);
	daemon_set_flag(DF_LOG_HOSTNAME);
	daemon_set_flag(DF_LOG_LL);
	if (conf.logfile_given) {
		daemon_set_logfile(conf.logfile_arg);
		daemon_open_log_or_die();
	}
	ret = parse_stream_args();
	if (ret < 0) {
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	log_welcome("para_audiod");
	server_uptime(UPTIME_SET);
	set_initial_status();
	FOR_EACH_SLOT(i)
		clear_slot(i);
	init_grabbing();
	setup_signal_handling();
	signal_setup_default(sig_task);

	init_status_task(stat_task);
	init_command_task(cmd_task);

	if (conf.daemon_given)
		daemonize();

	register_task(&sig_task->task);
	register_task(&cmd_task->task);
	register_task(&stat_task->task);
	s.default_timeout.tv_sec = 0;
	s.default_timeout.tv_usec = 99 * 1000;
	ret = schedule(&s);

	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	return EXIT_FAILURE;
}
