/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file udp_send.c Para_server's udp sender. */


#include <sys/time.h>
#include <dirent.h>
#include <sys/socket.h>
#include <net/if.h>

#include "server.cmdline.h"
#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "vss.h"
#include "list.h"
#include "send.h"
#include "portable_io.h"
#include "net.h"
#include "fd.h"
#include "sched.h"
#include "close_on_fork.h"
#include "chunk_queue.h"

/** Describes one entry in the list of targets for the udp sender. */
struct udp_target {
	/** The position of this target in the list of targets. */
	struct list_head node;
	/** The hostname (DNS name or IPv4/v6 address string). */
	char host[MAX_HOSTLEN];
	/** The UDP port. */
	int port;
	/** The socket fd. */
	int fd;
	/** The list of queued chunks for this fd. */
	struct chunk_queue *cq;
	/** The opaque structure returned by vss_add_fec_client(). */
	struct fec_client *fc;
	/** The FEC parameters for this target. */
	struct fec_client_parms fcp;
};

static struct list_head targets;
static int sender_status;

static void udp_close_target(struct udp_target *ut)
{
	if (ut->fd < 0)
		return;
	close(ut->fd);
	del_close_on_fork_list(ut->fd);
	cq_destroy(ut->cq);
	ut->cq = NULL;
	ut->fd = -1;
}

static void udp_delete_target(struct udp_target *ut, const char *msg)
{
	PARA_NOTICE_LOG("deleting %s#%d (%s) from list\n", ut->host,
		ut->port, msg);
	udp_close_target(ut);
	vss_del_fec_client(ut->fc);
	list_del(&ut->node);
	free(ut);
}

/**
 * Perform AF-independent multicast sender setup.
 *
 * \param fd	The connected socket descriptor.
 * \param ttl	UDPv4 multicast TTL or UDPv6 multicast number of hops.
 *		Use -1 to mean default, 0..255 otherwise.
 * \param iface	The outgoing multicast interface, or NULL for the default.
 *
 * \return Zero if okay, negative on error.
 */
static int mcast_sender_setup(struct udp_target *ut, int ttl, char *iface)
{
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof(ss);

	const int on = 1;
	int id = iface == NULL ? 0 : if_nametoindex(iface);

	if (getpeername(ut->fd, (struct sockaddr *)&ss, &sslen) < 0)
		goto err;

	if (iface != NULL && id == 0)
		PARA_WARNING_LOG("could not resolve interface %s, using default", iface);

	/* RFC 3493, 5.2: -1 means 'use kernel default' */
	if (ttl < 0 || ttl > 255)
		ttl = -1;

	switch (ss.ss_family) {
	case AF_INET:
		if (!IN_MULTICAST(htonl(((struct sockaddr_in *)&ss)->sin_addr.s_addr)))
			return 0;
		if (id != 0) {
#ifdef HAVE_IP_MREQN
			struct ip_mreqn mn;

			memset(&mn, 0, sizeof(mn));
			mn.imr_ifindex = id;
			if (setsockopt(ut->fd, IPPROTO_IP, IP_MULTICAST_IF, &mn, sizeof(mn)) < 0)
				goto err;
#else
			PARA_ERROR_LOG("No support for setting outgoing IPv4 mcast interface.");
#endif
		}
		/*
		 * Enable receiving multicast messages generated on the local host
		 * At least on Linux, this is enabled by default.
		 */
		if (setsockopt(ut->fd, IPPROTO_IP, IP_MULTICAST_LOOP, &on, sizeof(on)) < 0)
			break;

		/* Default: use local subnet (do not flood out into the WAN) */
		if (ttl == -1)
			ttl = 1;
		if (setsockopt(ut->fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0)
			break;
		return 0;
	case AF_INET6:
		if (!IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6 *)&ss)->sin6_addr))
			return 0;
		if (id != 0 &&
		    setsockopt(ut->fd, IPPROTO_IPV6, IPV6_MULTICAST_IF, &id, sizeof(id)) < 0)
			break;
		if (setsockopt(ut->fd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &on, sizeof(on)) < 0)
			break;
		if (setsockopt(ut->fd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &ttl, sizeof(ttl)) < 0)
			break;
		return 0;
	default:
		PARA_ERROR_LOG("address family %d not supported", ss.ss_family);
		return -E_ADDRESS_LOOKUP;
	}
err:
	return -ERRNO_TO_PARA_ERROR(errno);
}

/** The maximal size of the per-target chunk queue. */
#define UDP_CQ_BYTES 40000

static int udp_init_session(struct udp_target *ut)
{
	int ret;
	char *iface = NULL;

	if (ut->fd >= 0) /* nothing to do */
		return 0;

	ret = makesock(AF_UNSPEC, IPPROTO_UDP, 0, ut->host, ut->port);
	if (ret < 0)
		return ret;
	ut->fd = ret;

	if (conf.udp_mcast_iface_given)
		iface = conf.udp_mcast_iface_arg;

	ret = mcast_sender_setup(ut, conf.udp_ttl_arg, iface);
	if (ret < 0) {
		close(ut->fd);
		return ret;
	}

	ret = mark_fd_nonblocking(ut->fd);
	if (ret < 0) {
		close(ut->fd);
		return ret;
	}
	add_close_on_fork_list(ut->fd);
	ut->cq = cq_new(UDP_CQ_BYTES);
	PARA_NOTICE_LOG("sending to udp %s#%d using fec parms %d:%d:%d\n",
		ut->host, ut->port , ut->fcp.max_slice_bytes,
		ut->fcp.data_slices_per_group, ut->fcp.slices_per_group);
	return 1;
}

static void udp_shutdown_targets(void)
{
	struct udp_target *ut, *tmp;
	const char *buf = NULL;
	size_t len = 0; /* STFU, gcc */

	list_for_each_entry_safe(ut, tmp, &targets, node) {
		if (ut->fd < 0)
			continue;
		if (!buf)
			len = vss_get_fec_eof_packet(&buf);
		write(ut->fd, buf, len);
		udp_close_target(ut);
	}
}

static int udp_com_on(__a_unused struct sender_command_data *scd)
{
	sender_status = SENDER_ON;
	return 1;
}

static int udp_com_off(__a_unused struct sender_command_data *scd)
{
	udp_shutdown_targets();
	sender_status = SENDER_OFF;
	return 1;
}

static int udp_com_delete(struct sender_command_data *scd)
{
	struct udp_target *ut, *tmp;

	list_for_each_entry_safe(ut, tmp, &targets, node) {
		/* Unspecified port means wildcard port match */
		if (scd->port > 0 && scd->port != ut->port)
			continue;
		if (strcmp(ut->host, scd->host))
			continue;
		udp_delete_target(ut, "com_delete");
	}
	return 1;
}

static int udp_send_fec(char *buf, size_t len, void *private_data)
{
	struct udp_target *ut = private_data;
	int ret = udp_init_session(ut);

	if (ret < 0)
		goto fail;
	ret = send_queued_chunks(ut->fd, ut->cq, 0);
	if (ret < 0)
		goto fail;
	if (!len)
		return 0;
	if (!ret) { /* still data left in the queue */
		ret = cq_enqueue(ut->cq, buf, len);
		if (ret < 0)
			goto fail;
	}
	ret = write_nonblock(ut->fd, buf, len, 0);
	if (ret < 0)
		goto fail;
	if (ret != len) {
		ret = cq_enqueue(ut->cq, buf + ret, len - ret);
		if (ret < 0)
			goto fail;
	}
	return 1;
fail:
	udp_delete_target(ut, para_strerror(-ret));
	return ret;
}

static void udp_add_target(struct sender_command_data *scd)
{
	struct udp_target *ut = para_calloc(sizeof(struct udp_target));

	strncpy(ut->host, scd->host, sizeof(ut->host));
	ut->port = scd->port > 0 ? scd->port : conf.udp_default_port_arg;
	ut->fd = -1; /* not yet connected */
	PARA_INFO_LOG("adding to target list (%s#%d)\n", ut->host, ut->port);
	para_list_add(&ut->node, &targets);
	ut->fcp.slices_per_group = scd->slices_per_group;
	ut->fcp.data_slices_per_group = scd->data_slices_per_group;
	ut->fcp.max_slice_bytes = scd->max_slice_bytes;
	ut->fcp.send = udp_send_fec;
	ut->fcp.private_data = ut;
	vss_add_fec_client(&ut->fcp, &ut->fc);
}

static int udp_com_add(struct sender_command_data *scd)
{
	udp_add_target(scd);
	return 1;
}

static char *udp_info(void)
{
	struct udp_target *ut;
	char *ret, *tgts = NULL;

	list_for_each_entry(ut, &targets, node) {
		bool is_v6 = strchr(ut->host, ':') != NULL;
		char *tmp = make_message("%s%s%s%s:%d/%u:%u:%u ", tgts ? : "",
			is_v6 ? "[" : "", ut->host,
			is_v6 ? "]" : "", ut->port,
			ut->fcp.max_slice_bytes,
			ut->fcp.data_slices_per_group,
			ut->fcp.slices_per_group
		);
		free(tgts);
		tgts = tmp;
	}
	ret = make_message(
		"udp sender:\n"
		"\tstatus: %s\n"
		"\tport: udp %d\n"
		"\ttargets: %s\n",
		(sender_status == SENDER_ON)? "on" : "off",
		conf.udp_default_port_arg,
		tgts? tgts : "(none)"
	);
	free(tgts);
	return ret;
}

static void udp_init_target_list(void)
{
	struct sender_command_data scd;
	int i;

	INIT_LIST_HEAD(&targets);
	for (i = 0; i < conf.udp_target_given; i++) {
		if (parse_fec_url(conf.udp_target_arg[i], &scd) < 0) {
			PARA_CRIT_LOG("syntax error for udp target option "
				"#%d, ignoring\n", i);
			continue;
		}
		udp_add_target(&scd);
	}
}

static char *udp_help(void)
{
	return make_message(
		"usage: {on|off}\n"
		"usage: {add|delete} host[:port][/packet_size:k:n]\n"
		"examples: add 224.0.1.38:1500  (IPv4 multicast)\n"
		"          add 224.0.1.38:1500/1400:14:16\n"
		"              (likewise, using 1400 byte packets, with 14\n"
		"               data slices per 16 slice FEC group)\n"
		"          add 10.10.1.42       (using default port)\n"
		"          add [FF05::42]:1500  (IPv6 multicast)\n"
		"          add [::1]            (IPv6 localhost/default port)\n"
		"          delete myhost.net    (host with port wildcard)\n"
		"          delete [badc0de::1]  (IPv6 with port wildcard)\n"
	);
}

/**
 * The init function of para_server's udp sender.
 *
 * \param s Pointer to the udp sender struct.
 *
 * It initializes all function pointers of \a s and the list of udp targets.
 */
void udp_send_init(struct sender *s)
{
	INIT_LIST_HEAD(&targets);
	s->info = udp_info;
	s->help = udp_help;
	s->send = NULL;
	s->pre_select = NULL;
	s->post_select = NULL;
	s->shutdown_clients = udp_shutdown_targets;
	s->client_cmds[SENDER_ON] = udp_com_on;
	s->client_cmds[SENDER_OFF] = udp_com_off;
	s->client_cmds[SENDER_DENY] = NULL;
	s->client_cmds[SENDER_ALLOW] = NULL;
	s->client_cmds[SENDER_ADD] = udp_com_add;
	s->client_cmds[SENDER_DELETE] = udp_com_delete;
	sender_status = SENDER_OFF;
	udp_init_target_list();
	if (!conf.udp_no_autostart_given)
		sender_status = SENDER_ON;
	PARA_DEBUG_LOG("udp sender init complete\n");
}
